package ru.dbotthepony.kbox2d.collision.shapes

import ru.dbotthepony.kbox2d.api.*
import ru.dbotthepony.kvector.util2d.AABB
import ru.dbotthepony.kvector.vector.ndouble.Vector2d
import ru.dbotthepony.kvector.vector.ndouble.times
import kotlin.math.PI
import kotlin.math.sqrt

class CircleShape(
	override var radius: Double = 0.0,
	var position: Vector2d = Vector2d.ZERO,
) : IShape<CircleShape> {
	var p by this::position

	override fun copy(): CircleShape {
		return CircleShape().also {
			it.position = this.position
			it.radius = this.radius
		}
	}

	override val type: IShape.Type = IShape.Type.CIRCLE
	override val childCount: Int = 1

	override fun testPoint(transform: Transform, p: Vector2d): Boolean {
		val center = transform.p + b2Mul(transform.q, position)
		val d = p - center
		return b2Dot(d, d) <= radius * radius
	}

	override fun rayCast(input: RayCastInput, transform: Transform, childIndex: Int): RayCastOutput {
		val position = transform.p + b2Mul(transform.q, position)
		val s = input.p1 - position
		val b = b2Dot(s, s) - radius * radius

		// Solve quadratic equation.
		// Solve quadratic equation.
		val r = input.p2 - input.p1
		val c =  b2Dot(s, r)
		val rr = b2Dot(r, r)
		val sigma = c * c - rr * b

		// Check for negative discriminant and short segment.
		if (sigma < 0.0 || rr < b2_epsilon) {
			return RayCastOutput.MISS
		}

		// Find the point of intersection of the line with the circle.
		var a = -(c + sqrt(sigma))

		// Is the intersection point on the segment?
		if (a in 0.0 .. input.maxFraction * rr) {
			a /= rr
			return RayCastOutput(hit = true, fraction = a, normal = (s + a * r).normalized)
		}

		return RayCastOutput.MISS
	}

	override fun computeAABB(transform: Transform, childIndex: Int): AABB {
		val p = transform.p + b2Mul(transform.q, position)

		return AABB(
			mins = Vector2d(p.x - radius, p.y - radius),
			maxs = Vector2d(p.x + radius, p.y + radius)
		)
	}

	override fun computeMass(density: Double): MassData {
		val mass = density * PI * radius * radius

		return MassData(
			mass = mass,
			center = position,
			inertia = mass * (0.5 * radius * radius + b2Dot(position, position))
		)
	}
}
