
@file:Suppress("unused")

package ru.dbotthepony.kvector.api

import ru.dbotthepony.kvector.api.concrete.IMatrix2d
import java.nio.ByteBuffer
import java.nio.DoubleBuffer
import java.nio.FloatBuffer

interface IMatrixGetterDouble : IMatrixLike {
	/**
	 * @return component of this matrix, as double
	 */
	operator fun get(column: Int, row: Int): Double
}

interface IMatrixSetterDouble : IMatrixLike {
	/**
	 * Sets component of this matrix, as double
	 */
	operator fun set(column: Int, row: Int, value: Double)
}

/**
 * Defines methods applicable to matrices working with [Double]s
 */
interface IDoubleMatrix<T : IDoubleMatrix<T>> : IMatrixGetterDouble {
	/**
	 * Multiplies all matrix components by [other], returning new matrix as result
	 *
	 * @return new matrix
	 */
	operator fun times(other: Double): T

	/**
	 * Divides all matrix components by [other], returning new matrix as result
	 *
	 * @return new matrix
	 */
	operator fun div(other: Double): T

	/**
	 * Multiplies this and [other] matrices by general matrix multiplication rules,
	 * if they are compatible in dimensions, and return new matrix.
	 *
	 * @throws IllegalArgumentException if matrices are not applicable for multiplication
	 * @return new matrix
	 */
	operator fun times(other: IMatrixGetterDouble): T

	/**
	 * Adds components of both matrices, returning new matrix as result
	 *
	 * @throws IllegalArgumentException if matrices are different in dimensions
	 * @return new matrix
	 */
	operator fun plus(other: IMatrixGetterDouble): T

	/**
	 * Subtracts components of both matrices, returning new matrix as result
	 *
	 * @throws IllegalArgumentException if matrices are different in dimensions
	 * @return new matrix
	 */
	operator fun minus(other: IMatrixGetterDouble): T

	/**
	 * This matrix' trace. If matrix is not square matrix, null is returned.
	 */
	val trace: Double?

	/**
	 * This matrix' determinant. If matrix is not square matrix, null is returned.
	 */
	val determinant: Double?

	/**
	 * This matrix' inverse matrix (this * I = identity).
	 *
	 * This operation is computation intense, do cache its result.
	 * It is not recommended computing inverse matrix of matrices with order of 6 or bigger.
	 *
	 * If this matrix is not square matrix, or determinant is zero, null is returned.
	 */
	val inverse: T?

	/**
	 * This matrix' cofactor matrix.
	 *
	 * This operation is computation intense, do cache its result.
	 *
	 * If this matrix is not square matrix, null is returned.
	 */
	val cofactor: T?

	/**
	 * This matrix' adjugate matrix.
	 *
	 * This operation is computation intense, do cache its result.
	 *
	 * If this matrix is not square matrix, null is returned.
	 */
	val adjugate: T?

	/**
	 * Copies memory of this matrix to new [DoubleArray] in either [rowMajor] or column-major
	 * order.
	 *
	 * @return newly constructed [DoubleArray]
	 */
	fun toDoubleArray(rowMajor: Boolean = false): DoubleArray {
		val memory = DoubleArray(columns * rows)
		var i = 0

		if (rowMajor) {
			for (row in 0 until rows) {
				for (column in 0 until columns) {
					memory[i++] = this[column, row]
				}
			}
		} else {
			for (column in 0 until columns) {
				for (row in 0 until rows) {
					memory[i++] = this[column, row]
				}
			}
		}

		return memory
	}

	/**
	 * Writes this matrix to specified [output] at its current position
	 * in either [rowMajor] or column-major order.
	 */
	fun write(output: ByteBuffer, rowMajor: Boolean = false) {
		if (rowMajor) {
			for (row in 0 until rows) {
				for (column in 0 until columns) {
					output.putDouble(this[column, row])
				}
			}
		} else {
			for (column in 0 until columns) {
				for (row in 0 until rows) {
					output.putDouble(this[column, row])
				}
			}
		}
	}

	/**
	 * Writes this matrix to specified [output] at its current position
	 * in either [rowMajor] or column-major order.
	 */
	fun write(output: DoubleBuffer, rowMajor: Boolean = false) {
		if (rowMajor) {
			for (row in 0 until rows) {
				for (column in 0 until columns) {
					output.put(this[column, row])
				}
			}
		} else {
			for (column in 0 until columns) {
				for (row in 0 until rows) {
					output.put(this[column, row])
				}
			}
		}
	}
}

interface IMutableDoubleMatrix<T : IMutableDoubleMatrix<T>> : IMatrixSetterDouble {
	/**
	 * Multiplies all matrix components by [other], writing result into this matrix
	 *
	 * @return this matrix
	 */
	fun timesMut(other: Double): T

	/**
	 * Divides all matrix components by [other], writing result into this matrix
	 *
	 * @return this matrix
	 */
	fun divMut(other: Double): T

	/**
	 * Multiplies this and [other] matrices by general matrix multiplication rules,
	 * if they are compatible in dimensions, and write result into this matrix.
	 *
	 * @throws IllegalArgumentException if [other] is not a square matrix
	 * @throws IllegalStateException if this matrix is not a square matrix
	 * @return this matrix
	 */
	fun timesMut(other: IMatrixGetterDouble): T

	/**
	 * Adds components of both matrices, writing result into this matrix
	 *
	 * @throws IllegalArgumentException if matrices are different in dimensions
	 * @return this matrix
	 */
	fun plusMut(other: IMatrixGetterDouble): T

	/**
	 * Subtracts components of both matrices, writing result into this matrix
	 *
	 * @throws IllegalArgumentException if matrices are different in dimensions
	 * @return this matrix
	 */
	fun minusMut(other: IMatrixGetterDouble): T
}
