
@file:Suppress("unused")

package ru.dbotthepony.kvector.api

import java.nio.ByteBuffer
import java.nio.FloatBuffer

interface IMatrixGetterFloat : IMatrixLike {
	/**
	 * @return component of this matrix, as float
	 */
	operator fun get(column: Int, row: Int): Float
}

interface IMatrixSetterFloat : IMatrixLike {
	/**
	 * Sets component of this matrix, as float
	 */
	operator fun set(column: Int, row: Int, value: Float)
}

/**
 * Defines methods applicable to matrices working with [Float]s
 */
interface IFloatMatrix<T : IFloatMatrix<T>> : IMatrixGetterFloat {
	/**
	 * Multiplies all matrix components by [other], returning new matrix as result
	 *
	 * @return new matrix
	 */
	operator fun times(other: Float): T

	/**
	 * Divides all matrix components by [other], returning new matrix as result
	 *
	 * @return new matrix
	 */
	operator fun div(other: Float): T

	/**
	 * Multiplies this and [other] matrices by general matrix multiplication rules,
	 * if they are compatible in dimensions, and return new matrix.
	 *
	 * @throws IllegalArgumentException if matrices are not applicable for multiplication
	 * @return new matrix
	 */
	operator fun times(other: IMatrixGetterFloat): T

	/**
	 * Adds components of both matrices, returning new matrix as result
	 *
	 * @throws IllegalArgumentException if matrices are different in dimensions
	 * @return new matrix
	 */
	operator fun plus(other: IMatrixGetterFloat): T

	/**
	 * Subtracts components of both matrices, returning new matrix as result
	 *
	 * @throws IllegalArgumentException if matrices are different in dimensions
	 * @return new matrix
	 */
	operator fun minus(other: IMatrixGetterFloat): T

	/**
	 * This matrix trace. If matrix is not square matrix, null is returned.
	 */
	val trace: Float?

	/**
	 * This matrix determinant. If matrix is not square matrix, null is returned.
	 */
	val determinant: Float?

	/**
	 * This matrix' inverse matrix (this * I = identity).
	 *
	 * This operation is computation intense, do cache its result.
	 *
	 * If this matrix is not square matrix, or determinant is zero, null is returned.
	 */
	val inverse: T?

	/**
	 * This matrix' cofactor matrix.
	 *
	 * This operation is computation intense, do cache its result.
	 *
	 * If this matrix is not square matrix, null is returned.
	 */
	val cofactor: T?

	/**
	 * This matrix' adjugate matrix.
	 *
	 * This operation is computation intense, do cache its result.
	 *
	 * If this matrix is not square matrix, null is returned.
	 */
	val adjugate: T?

	/**
	 * Copies memory of this matrix to new [FloatArray] in either [rowMajor] or column-major
	 * order.
	 *
	 * @return newly constructed [FloatArray]
	 */
	fun toFloatArray(rowMajor: Boolean = false): FloatArray {
		val memory = FloatArray(columns * rows)
		var i = 0

		if (rowMajor) {
			for (row in 0 until rows) {
				for (column in 0 until columns) {
					memory[i++] = this[column, row]
				}
			}
		} else {
			for (column in 0 until columns) {
				for (row in 0 until rows) {
					memory[i++] = this[column, row]
				}
			}
		}

		return memory
	}

	/**
	 * Writes this matrix to specified [output] at its current position
	 * in either [rowMajor] or column-major order.
	 */
	fun write(output: ByteBuffer, rowMajor: Boolean = false) {
		if (rowMajor) {
			for (row in 0 until rows) {
				for (column in 0 until columns) {
					output.putFloat(this[column, row])
				}
			}
		} else {
			for (column in 0 until columns) {
				for (row in 0 until rows) {
					output.putFloat(this[column, row])
				}
			}
		}
	}

	/**
	 * Writes this matrix to specified [output] at its current position
	 * in either [rowMajor] or column-major order.
	 */
	fun write(output: FloatBuffer, rowMajor: Boolean = false) {
		if (rowMajor) {
			for (row in 0 until rows) {
				for (column in 0 until columns) {
					output.put(this[column, row])
				}
			}
		} else {
			for (column in 0 until columns) {
				for (row in 0 until rows) {
					output.put(this[column, row])
				}
			}
		}
	}
}

interface IMutableFloatMatrix<T : IMutableFloatMatrix<T>> : IMatrixSetterFloat {
	/**
	 * Multiplies all matrix components by [other], writing result into this matrix
	 *
	 * @return this matrix
	 */
	fun timesMut(other: Float): T

	/**
	 * Divides all matrix components by [other], writing result into this matrix
	 *
	 * @return this matrix
	 */
	fun divMut(other: Float): T

	/**
	 * Multiplies this and [other] matrices by general matrix multiplication rules,
	 * if they are compatible in dimensions, and write result into this matrix.
	 *
	 * @throws IllegalArgumentException if [other] is not a square matrix
	 * @throws IllegalStateException if this matrix is not a square matrix
	 * @return this matrix
	 */
	fun timesMut(other: IMatrixGetterFloat): T

	/**
	 * Adds components of both matrices, writing result into this matrix
	 *
	 * @throws IllegalArgumentException if matrices are different in dimensions
	 * @return this matrix
	 */
	fun plusMut(other: IMatrixGetterFloat): T

	/**
	 * Subtracts components of both matrices, writing result into this matrix
	 *
	 * @throws IllegalArgumentException if matrices are different in dimensions
	 * @return this matrix
	 */
	fun minusMut(other: IMatrixGetterFloat): T
}
