package ru.dbotthepony.kvector.api.concrete

import ru.dbotthepony.kvector.api.IDoubleMatrix
import ru.dbotthepony.kvector.api.IMatrix
import ru.dbotthepony.kvector.api.ISquareMatrix
import ru.dbotthepony.kvector.api.ITransposable
import ru.dbotthepony.kvector.matrix.*
import ru.dbotthepony.kvector.matrix.ndouble.*
import ru.dbotthepony.kvector.matrix.nfloat.*
import ru.dbotthepony.kvector.vector.ndouble.Vector2d
import ru.dbotthepony.kvector.vector.ndouble.Vector3d
import ru.dbotthepony.kvector.vector.ndouble.Vector4d

/**
 * [Matrix2d] and [MutableMatrix2d] implement this
 */
interface IMatrix2d<T : IMatrix2d<T>> : IMatrix<T>, IDoubleMatrix<T>, ITransposable<T> {
	val c00: Double
	val c10: Double
	val c01: Double
	val c11: Double
	val r00: Double
	val r01: Double
	val r10: Double
	val r11: Double
	val a11: Double
	val a21: Double
	val a12: Double
	val a22: Double
	val b11: Double
	val b21: Double
	val b12: Double
	val b22: Double
	val c0: Vector2d
	val c1: Vector2d
	val r0: Vector2d
	val r1: Vector2d

	/**
	 * Creates a copy of this matrix as [Matrix4d],
	 * setting last two rows and columns to 0
	 */
	fun toMatrix4d(): Matrix4d {
		return Matrix4d(
			c00, c01, 0.0, 0.0,
			c10, c11, 0.0, 0.0,
			0.0, 0.0, 0.0, 0.0,
			0.0, 0.0, 0.0, 0.0,
		)
	}

	/**
	 * Creates a copy of this matrix as [Matrix3d],
	 * setting last row and column to 0
	 */
	fun toMatrix3d(): Matrix3d {
		return Matrix3d(
			c00, c01, 0.0,
			c10, c11, 0.0,
			0.0, 0.0, 0.0,
		)
	}

	/**
	 * Creates a copy of this matrix as [Matrix2d]
	 */
	fun toMatrix2d(): Matrix2d {
		return Matrix2d(
			c00, c01,
			c10, c11,
		)
	}

	/**
	 * Creates a copy of this matrix as [MutableMatrix4d],
	 * setting last two columns and rows to 0
	 */
	fun toMutableMatrix4d(): MutableMatrix4d {
		return MutableMatrix4d(
			c00, c01, 0.0, 0.0,
			c10, c11, 0.0, 0.0,
			0.0, 0.0, 0.0, 0.0,
			0.0, 0.0, 0.0, 0.0,
		)
	}

	/**
	 * Creates a copy of this matrix as [MutableMatrix3d],
	 * setting last column and row to 0
	 */
	fun toMutableMatrix3d(): MutableMatrix3d {
		return MutableMatrix3d(
			c00, c01, 0.0,
			c10, c11, 0.0,
			0.0, 0.0, 0.0,
		)
	}

	/**
	 * Creates a copy of this matrix as [MutableMatrix2d]
	 */
	fun toMutableMatrix2d(): MutableMatrix2d {
		return MutableMatrix2d(
			c00, c01,
			c10, c11,
		)
	}

	/**
	 * Creates a copy of this matrix as [Matrix4f],
	 * casing all [Double]s to [Float]s,
	 * setting last two columns and rows to 0
	 */
	fun toMatrix4f(): Matrix4f {
		return Matrix4f(
			c00.toFloat(), c01.toFloat(), 0f, 0f,
			c10.toFloat(), c11.toFloat(), 0f, 0f,
			0f, 0f, 0f, 0f,
			0f, 0f, 0f, 0f,
		)
	}

	/**
	 * Creates a copy of this matrix as [Matrix3f],
	 * casing all [Double]s to [Float]s,
	 * setting last column and row to 0
	 */
	fun toMatrix3f(): Matrix3f {
		return Matrix3f(
			c00.toFloat(), c01.toFloat(), 0f,
			c10.toFloat(), c11.toFloat(), 0f,
			0f, 0f, 0f,
		)
	}

	/**
	 * Creates a copy of this matrix as [Matrix2f],
	 * truncating last row and column
	 */
	fun toMatrix2f(): Matrix2f {
		return Matrix2f(
			c00.toFloat(), c01.toFloat(),
			c10.toFloat(), c11.toFloat(),
		)
	}

	/**
	 * Creates a copy of this matrix as [MutableMatrix4f],
	 * setting last two columns and rows to 0,
	 * casing all [Double]s to [Float]s
	 */
	fun toMutableMatrix4f(): MutableMatrix4f {
		return MutableMatrix4f(
			c00.toFloat(), c01.toFloat(), 0f, 0f,
			c10.toFloat(), c11.toFloat(), 0f, 0f,
			0f, 0f, 0f, 0f,
			0f, 0f, 0f, 0f,
		)
	}

	/**
	 * Creates a copy of this matrix as [MutableMatrix3f],
	 * setting last column and row to 0,
	 * casing all [Double]s to [Float]s
	 */
	fun toMutableMatrix3f(): MutableMatrix3f {
		return MutableMatrix3f(
			c00.toFloat(), c01.toFloat(), 0f,
			c10.toFloat(), c11.toFloat(), 0f,
			0f, 0f, 0f,
		)
	}

	/**
	 * Creates a copy of this matrix as [MutableMatrix2f],
	 * truncating last row and column,
	 * casing all [Double]s to [Float]s
	 */
	fun toMutableMatrix2f(): MutableMatrix2f {
		return MutableMatrix2f(
			c00.toFloat(), c01.toFloat(),
			c10.toFloat(), c11.toFloat(),
		)
	}
}

/**
 * [Matrix3d] and [MutableMatrix3d] implement this
 */
interface IMatrix3d<T : IMatrix3d<T>> : IMatrix<T>, IDoubleMatrix<T>, ITransposable<T>, ISquareMatrix<T, Vector2d, Vector3d> {
	val c00: Double
	val c10: Double
	val c20: Double
	val c01: Double
	val c11: Double
	val c21: Double
	val c02: Double
	val c12: Double
	val c22: Double
	val r00: Double
	val r01: Double
	val r02: Double
	val r10: Double
	val r11: Double
	val r12: Double
	val r20: Double
	val r21: Double
	val r22: Double
	val a11: Double
	val a21: Double
	val a31: Double
	val a12: Double
	val a22: Double
	val a32: Double
	val a13: Double
	val a23: Double
	val a33: Double
	val b11: Double
	val b21: Double
	val b31: Double
	val b12: Double
	val b22: Double
	val b32: Double
	val b13: Double
	val b23: Double
	val b33: Double
	val c0: Vector3d
	val c1: Vector3d
	val c2: Vector3d
	val r0: Vector3d
	val r1: Vector3d
	val r2: Vector3d

	override val translation: Vector2d get() = Vector2d(r02, r12)

	/**
	 * Creates a copy of this matrix as [Matrix4d],
	 * setting last row and column to 0
	 */
	fun toMatrix4d(): Matrix4d {
		return Matrix4d(
			c00, c01, c02, 0.0,
			c10, c11, c12, 0.0,
			c20, c21, c22, 0.0,
			0.0, 0.0, 0.0, 0.0,
		)
	}

	/**
	 * Creates a copy of this matrix as [Matrix3d],
	 */
	fun toMatrix3d(): Matrix3d {
		return Matrix3d(
			c00, c01, c02,
			c10, c11, c12,
			c20, c21, c22,
		)
	}

	/**
	 * Creates a copy of this matrix as [Matrix2d],
	 * truncating last row and column
	 */
	fun toMatrix2d(): Matrix2d {
		return Matrix2d(
			c00, c01,
			c10, c11,
		)
	}

	/**
	 * Creates a copy of this matrix as [MutableMatrix4d],
	 * setting last column and row to 0
	 */
	fun toMutableMatrix4d(): MutableMatrix4d {
		return MutableMatrix4d(
			c00, c01, c02, 0.0,
			c10, c11, c12, 0.0,
			c20, c21, c22, 0.0,
			0.0, 0.0, 0.0, 0.0,
		)
	}

	/**
	 * Creates a copy of this matrix as [MutableMatrix3d]
	 */
	fun toMutableMatrix3d(): MutableMatrix3d {
		return MutableMatrix3d(
			c00, c01, c02,
			c10, c11, c12,
			c20, c21, c22,
		)
	}

	/**
	 * Creates a copy of this matrix as [MutableMatrix2d],
	 * truncating last row and column
	 */
	fun toMutableMatrix2d(): MutableMatrix2d {
		return MutableMatrix2d(
			c00, c01,
			c10, c11,
		)
	}

	/**
	 * Creates a copy of this matrix as [Matrix4f],
	 * casing all [Double]s to [Float]s,
	 * setting last column and row to 0
	 */
	fun toMatrix4f(): Matrix4f {
		return Matrix4f(
			c00.toFloat(), c01.toFloat(), c02.toFloat(), 0f,
			c10.toFloat(), c11.toFloat(), c12.toFloat(), 0f,
			c20.toFloat(), c21.toFloat(), c22.toFloat(), 0f,
			0f, 0f, 0f, 0f,
		)
	}

	/**
	 * Creates a copy of this matrix as [Matrix3f],
	 * casing all [Double]s to [Float]s
	 */
	fun toMatrix3f(): Matrix3f {
		return Matrix3f(
			c00.toFloat(), c01.toFloat(), c02.toFloat(),
			c10.toFloat(), c11.toFloat(), c12.toFloat(),
			c20.toFloat(), c21.toFloat(), c22.toFloat(),
		)
	}

	/**
	 * Creates a copy of this matrix as [Matrix2f],
	 * truncating last row and column,
	 * casing all [Double]s to [Float]s
	 */
	fun toMatrix2f(): Matrix2f {
		return Matrix2f(
			c00.toFloat(), c01.toFloat(),
			c10.toFloat(), c11.toFloat(),
		)
	}

	/**
	 * Creates a copy of this matrix as [MutableMatrix4f],
	 * setting last column and row to 0,
	 * casing all [Double]s to [Float]s
	 */
	fun toMutableMatrix4f(): MutableMatrix4f {
		return MutableMatrix4f(
			c00.toFloat(), c01.toFloat(), c02.toFloat(), 0f,
			c10.toFloat(), c11.toFloat(), c12.toFloat(), 0f,
			c20.toFloat(), c21.toFloat(), c22.toFloat(), 0f,
			0f, 0f, 0f, 0f,
		)
	}

	/**
	 * Creates a copy of this matrix as [MutableMatrix3f],
	 * casing all [Double]s to [Float]s
	 */
	fun toMutableMatrix3f(): MutableMatrix3f {
		return MutableMatrix3f(
			c00.toFloat(), c01.toFloat(), c02.toFloat(),
			c10.toFloat(), c11.toFloat(), c12.toFloat(),
			c20.toFloat(), c21.toFloat(), c22.toFloat(),
		)
	}

	/**
	 * Creates a copy of this matrix as [MutableMatrix2f],
	 * truncating last row and column,
	 * casing all [Double]s to [Float]s
	 */
	fun toMutableMatrix2f(): MutableMatrix2f {
		return MutableMatrix2f(
			c00.toFloat(), c01.toFloat(),
			c10.toFloat(), c11.toFloat(),
		)
	}
}

/**
 * [Matrix4d] and [MutableMatrix4d] implement this
 */
interface IMatrix4d<T : IMatrix4d<T>> : IMatrix<T>, IDoubleMatrix<T>, ITransposable<T>, ISquareMatrix<T, Vector3d, Vector4d> {
	val c00: Double
	val c10: Double
	val c20: Double
	val c30: Double
	val c01: Double
	val c11: Double
	val c21: Double
	val c31: Double
	val c02: Double
	val c12: Double
	val c22: Double
	val c32: Double
	val c03: Double
	val c13: Double
	val c23: Double
	val c33: Double
	val r00: Double
	val r01: Double
	val r02: Double
	val r03: Double
	val r10: Double
	val r11: Double
	val r12: Double
	val r13: Double
	val r20: Double
	val r21: Double
	val r22: Double
	val r23: Double
	val r30: Double
	val r31: Double
	val r32: Double
	val r33: Double
	val a11: Double
	val a21: Double
	val a31: Double
	val a41: Double
	val a12: Double
	val a22: Double
	val a32: Double
	val a42: Double
	val a13: Double
	val a23: Double
	val a33: Double
	val a43: Double
	val a14: Double
	val a24: Double
	val a34: Double
	val a44: Double
	val b11: Double
	val b21: Double
	val b31: Double
	val b41: Double
	val b12: Double
	val b22: Double
	val b32: Double
	val b42: Double
	val b13: Double
	val b23: Double
	val b33: Double
	val b43: Double
	val b14: Double
	val b24: Double
	val b34: Double
	val b44: Double
	val c0: Vector4d
	val c1: Vector4d
	val c2: Vector4d
	val c3: Vector4d
	val r0: Vector4d
	val r1: Vector4d
	val r2: Vector4d
	val r3: Vector4d

	override val translation: Vector3d get() = Vector3d(r03, r13, r23)

	/**
	 * Creates a copy of this matrix as [Matrix4d]
	 */
	fun toMatrix4d(): Matrix4d {
		return Matrix4d(
			c00, c01, c02, c03,
			c10, c11, c12, c13,
			c20, c21, c22, c23,
			c30, c31, c32, c33,
		)
	}

	/**
	 * Creates a copy of this matrix as [Matrix3d],
	 * truncating last row and column
	 */
	fun toMatrix3d(): Matrix3d {
		return Matrix3d(
			c00, c01, c02,
			c10, c11, c12,
			c20, c21, c22,
		)
	}

	/**
	 * Creates a copy of this matrix as [Matrix2d],
	 * truncating two last rows and columns
	 */
	fun toMatrix2d(): Matrix2d {
		return Matrix2d(
			c00, c01,
			c10, c11,
		)
	}

	/**
	 * Creates a copy of this matrix as [MutableMatrix4d]
	 */
	fun toMutableMatrix4d(): MutableMatrix4d {
		return MutableMatrix4d(
			c00, c01, c02, c03,
			c10, c11, c12, c13,
			c20, c21, c22, c23,
			c30, c31, c32, c33,
		)
	}

	/**
	 * Creates a copy of this matrix as [MutableMatrix3d],
	 * truncating last row and column
	 */
	fun toMutableMatrix3d(): MutableMatrix3d {
		return MutableMatrix3d(
			c00, c01, c02,
			c10, c11, c12,
			c20, c21, c22,
		)
	}

	/**
	 * Creates a copy of this matrix as [MutableMatrix2d],
	 * truncating two last rows and columns
	 */
	fun toMutableMatrix2d(): MutableMatrix2d {
		return MutableMatrix2d(
			c00, c01,
			c10, c11,
		)
	}

	/**
	 * Creates a copy of this matrix as [Matrix4f],
	 * casing all [Double]s to [Float]s
	 */
	fun toMatrix4f(): Matrix4f {
		return Matrix4f(
			c00.toFloat(), c01.toFloat(), c02.toFloat(), c03.toFloat(),
			c10.toFloat(), c11.toFloat(), c12.toFloat(), c13.toFloat(),
			c20.toFloat(), c21.toFloat(), c22.toFloat(), c23.toFloat(),
			c30.toFloat(), c31.toFloat(), c32.toFloat(), c33.toFloat(),
		)
	}

	/**
	 * Creates a copy of this matrix as [Matrix3f],
	 * truncating last row and column,
	 * casing all [Double]s to [Float]s
	 */
	fun toMatrix3f(): Matrix3f {
		return Matrix3f(
			c00.toFloat(), c01.toFloat(), c02.toFloat(),
			c10.toFloat(), c11.toFloat(), c12.toFloat(),
			c20.toFloat(), c21.toFloat(), c22.toFloat(),
		)
	}

	/**
	 * Creates a copy of this matrix as [Matrix2f],
	 * truncating two last rows and columns,
	 * casing all [Double]s to [Float]s
	 */
	fun toMatrix2f(): Matrix2f {
		return Matrix2f(
			c00.toFloat(), c01.toFloat(),
			c10.toFloat(), c11.toFloat(),
		)
	}

	/**
	 * Creates a copy of this matrix as [MutableMatrix4f],
	 * casing all [Double]s to [Float]s
	 */
	fun toMutableMatrix4f(): MutableMatrix4f {
		return MutableMatrix4f(
			c00.toFloat(), c01.toFloat(), c02.toFloat(), c03.toFloat(),
			c10.toFloat(), c11.toFloat(), c12.toFloat(), c13.toFloat(),
			c20.toFloat(), c21.toFloat(), c22.toFloat(), c23.toFloat(),
			c30.toFloat(), c31.toFloat(), c32.toFloat(), c33.toFloat(),
		)
	}

	/**
	 * Creates a copy of this matrix as [MutableMatrix3f],
	 * truncating last row and column,
	 * casing all [Double]s to [Float]s
	 */
	fun toMutableMatrix3f(): MutableMatrix3f {
		return MutableMatrix3f(
			c00.toFloat(), c01.toFloat(), c02.toFloat(),
			c10.toFloat(), c11.toFloat(), c12.toFloat(),
			c20.toFloat(), c21.toFloat(), c22.toFloat(),
		)
	}

	/**
	 * Creates a copy of this matrix as [MutableMatrix2f],
	 * truncating two last rows and columns,
	 * casing all [Double]s to [Float]s
	 */
	fun toMutableMatrix2f(): MutableMatrix2f {
		return MutableMatrix2f(
			c00.toFloat(), c01.toFloat(),
			c10.toFloat(), c11.toFloat(),
		)
	}
}
